#!usr/bin/env python
"""
Calculate various summary statistics for samples placed in
respective OTUs. User inputs minimum number of alleles
required for stats to be calculated for a population.
Takes infile.alleles file from pyrad output as input data.
Requires python package dendropy for calculations.

author: J. Satler
date: 10 July 2018
version: 2

python pyNucDiv.py Traits.file Infile.allele alleles_threshold
"""

import os
import sys
import shutil
import numpy as np
import dendropy
from dendropy.calculate import popgenstat

def get_otus(file):
    """assign individuals to OTUs"""
    with open(file, 'r') as p:
        return {line.split()[0].strip():line.split()[1].strip()
               for line in p if not line.startswith("Ind\t")}

def all_otus(file):
    """get list of OTUs"""
    pops = []
    for k, v in file.items():
        if v not in pops:
            pops.append(v)
    return pops

def break_loci(file, pops, otuD, thr):
    with open(file, 'r') as f:
        #Breaks file at each '//' signalling a new locus.
        pops = {i:[] for i in pops}

        locus = 1
        for line in f:
            line = line.strip().split()
            if '//' not in line[0]:
                #Place individual in proper population
                pops[otuD[line[0][1:]]].append(line[0] + '\n'
                                               + line[1] + '\n')
            else:
                for k,v in pops.items():
                    #place sampling thresholds here
                    if len(v) >= int(thr):
                        writeO(''.join(v), locus, k)
                locus += 1
                #reset dictionary
                pops = {i:[] for i in pops}

def writeO(file, locus, pop):
    #Writes the locus to output with an arbitrary unique name
    with open(pop + '-' + str(locus) + '.fasta', 'w') as w:
        w.write(file)

def batch_calc(Infiles, pops):
    """generate summary statistics for OTUs"""
    pops = {i:[] for i in pops}

    fa = [f for f in os.listdir('./') if f.endswith('fasta')]
    for i in fa:
        o = get_ss(i)
        pops[i.split('-')[0]].append(o)
    return pops

def get_ss(fasta):
    """read loci and return list of summary statistics"""
    seqs = dendropy.DnaCharacterMatrix.get(path = fasta,
                                           schema = 'fasta')

    #test if Tajima's D can be calculated
    try:
        popgenstat.tajimas_d(seqs)
    except ZeroDivisionError:
        #print "Tajima's D gave a zero division error!"
        return [popgenstat.nucleotide_diversity(seqs),
                popgenstat.num_segregating_sites(seqs),
                popgenstat.wattersons_theta(seqs)]
    #return all four SS if Tajima's D calculated
    return [popgenstat.nucleotide_diversity(seqs),
            popgenstat.num_segregating_sites(seqs),
            popgenstat.wattersons_theta(seqs),
            popgenstat.tajimas_d(seqs)]

def put_ss_together(ssList):
    """place summary statistics into separate lists"""
    stats = {}

    for k, v in ssList.items():
        pi = [l[0] for l in v]
        S = [l[1] for l in v]
        Theta = [l[2] for l in v]
        D = [l[3] for l in v if len(l) == 4]
        #add SS to new dictionary
        #stats[k] = [pi, S, Theta]
        stats[k] = [pi, S, Theta, D]

    return stats

def write_ss(ss_pop):
    """write summary statistics to a file"""
    outfile = "Results_PopGen_{0}_thr{1}.txt".format(sys.argv[2].split('_')[0],
                                                     sys.argv[3])
    with open(outfile, 'a') as out:
        out.write("__Infiles__\nAlleles:\t{0}\nTraits:\t{1}\n \
                   \nMin # of alleles per pop:\t{2}\n\n\n".format(sys.argv[2],
                                                                  sys.argv[1],
                                                                  sys.argv[3]))
        for pop, stats in sorted(ss_pop.items()):
            out.write('***' + pop + '***\n\n')
            for i in range(len(stats)):
                if i == 0:
                    out.write("Pi\tN\t{0}\tmean\t{1:.4f}\tsd\t{2:.4f}\n".format(
                               len(stats[i]), np.mean(stats[i]), np.std(stats[i])))
                elif i == 1:
                    out.write("SS\tN\t{0}\tmean\t{1:.4f}\tsd\t{2:.4f}\n".format(
                               len(stats[i]), np.mean(stats[i]), np.std(stats[i])))
                elif i == 2:
                    out.write("Watterson's theta\tN\t{0}\tmean\t{1:.4f}\tsd\t{2:.4f}\n".format(
                               len(stats[i]), np.mean(stats[i]), np.std(stats[i])))
                else:
                    out.write("Tajima's D\tN\t{0}\tmean\t{1:.4f}\tsd\t{2:.4f}\n\n\n".format(
                               len(stats[i]), np.mean(stats[i]), np.std(stats[i])))

def move_files(pops):
    """move fasta files to results folder"""
    for p in pops:
        os.makedirs('./Results/' + p)
        fa = [shutil.move(f, './Results/' + p + '/' + f) for f in os.listdir('./')
                                                         if f.endswith('fasta')
                                                         if f.split('-')[0] == p]

def Main():
    if len(sys.argv) != 4:
        print "python pyNucDiv.py Traits.file Infile.allele alleles_threshold"
        sys.exit()
    OTUs = get_otus(sys.argv[1])
    pops = all_otus(OTUs)
    files = break_loci(sys.argv[2], pops, OTUs, sys.argv[3])
    print "Calculating SS now..."
    sumstats = batch_calc(files, pops)
    print "Combine SS into like lists..."
    ss_list = put_ss_together(sumstats)
    write_ss(ss_list)
    move_files(pops)

if __name__ == '__main__':
    Main()
